<?php

namespace Drp\NativeEventApp\Models;

use Carbon\Carbon;
use Drp\NativeEventApp\Database\TableName;
use Drp\NativeEventApp\Enums\NotificationStatus;
use Drp\NativeEventApp\Exceptions\NotificationNotScheduled;
use Drp\NativeEventApp\Models\Concerns\PrefixedTable;
use Drp\NativeEventApp\Models\Concerns\Segmentable;
use Drp\NativeEventApp\Templates\Compiler;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Database\Eloquent\Relations\BelongsToMany;
use Illuminate\Support\Collection;
use InvalidArgumentException;

class Notification extends Model
{
    use Segmentable, PrefixedTable;

    /** @var string */
    protected $table = 'notifications';

    /** @var array */
    protected $fillable = [
        'event_id',
        'title',
        'message',
        'status',
        'send_at',
        'sent_at',
        'total_sent',
    ];

    /** @var array */
    protected $dates = [
        'send_at',
        'sent_at',
    ];

    public static function make(Event $event, string $title, string $message): self
    {
        return new self([
            'event_id' => $event->id,
            'title' => $title,
            'message' => $message,
        ]);
    }

    public function event(): BelongsTo
    {
        return $this->belongsTo(Event::class);
    }

    public function devices(): BelongsToMany
    {
        return $this->belongsToMany(
            Device::class,
            TableName::make('device_notifications')
        );
    }

    public function forSegment(Segment $segment): self
    {
        return tap($this, static function (self $notification) use ($segment): void {
            $notification->ensureModelExists()->joinSegments(collect([$segment]));
        });
    }

    public function forSegments(Collection $segments): self
    {
        return tap($this, static function (self $notification) use ($segments): void {
            $notification->ensureModelExists()->joinSegments($segments);
        });
    }

    public function forDevice(Device $device): self
    {
        return tap($this, static function (self $notification) use ($device): void {
            $notification->ensureModelExists()->devices()->save($device);
        });
    }

    public function sendAt(Carbon $sendAt): self
    {
        throw_if(
            $sendAt->isPast(),
            new InvalidArgumentException('The send at date must not be in the past')
        );

        $this->send_at = $sendAt;

        return $this;
    }

    public function sendNow(): self
    {
        return $this->sendAt(now())->publish();
    }

    public function publish(): self
    {
        throw_if($this->send_at === null, NotificationNotScheduled::class);

        return tap($this, static function (self $notification): void {
            $notification->status = NotificationStatus::PUBLISHED;
            $notification->save();
        });
    }

    public function unpublish(): self
    {
        return tap($this, static function (self $notification): void {
            $notification->status = NotificationStatus::DRAFT;
            $notification->save();
        });
    }

    public function unschedule(): self
    {
        return tap($this, static function (self $notification): void {
            $notification->send_at = null;
            $notification->save();
        });
    }

    public function totalSent(int $totalSent): self
    {
        return tap($this, static function (self $notification) use ($totalSent): void {
            $notification->total_sent = $totalSent;
            $notification->save();
        });
    }

    public function ensureModelExists(): self
    {
        if ($this->exists) {
            return $this;
        }

        return tap($this)->save();
    }

    /**
     * @return self|Collection
     */
    public static function get(?int $notificationId = null)
    {
        if ($notificationId !== null) {
            return self::findOrFail($notificationId);
        }

        return self::all();
    }

    public static function getBySegment(Segment $segment): Collection
    {
        return self::hasSegment($segment)->get();
    }

    public function scopePublished(Builder $query): Builder
    {
        return $query->where('status', NotificationStatus::PUBLISHED);
    }

    public function scopeUnpublished(Builder $query): Builder
    {
        return $query->where('status', NotificationStatus::PUBLISHED);
    }

    public function scopeUnsent(Builder $query): Builder
    {
        return $query->where('sent_at', null)
            ->whereNotNull('send_at')
            ->published();
    }

    public function scopeShouldSend(Builder $query): Builder
    {
        return $query->where('send_at', '<=', now());
    }

    public function scopeSent(Builder $query): Builder
    {
        return $query->whereNotNull('sent_at');
    }

    public function scopeToBeSent(Builder $query): Builder
    {
        return $query->unsent()->shouldSend();
    }

    public static function getUnsent(): Collection
    {
        return self::unsent()->get();
    }

    public static function getUnsentWithoutSegment(): Collection
    {
        return self::unsent()
            ->notSegmented()
            ->get();
    }

    public static function getSent(): Collection
    {
        return self::sent()->get();
    }

    public function delete(): ?bool
    {
        $this->devices()->detach();
        $this->segments()->detach();

        return parent::delete();
    }

    public function markAsSent(): self
    {
        $this->sent_at = now();

        return tap($this)->save();
    }

    public function renderTitle(Registration $registration): string
    {
        return app(Compiler::class)
            ->for($registration)
            ->render($this->title);
    }

    public function renderMessage(Registration $registration): string
    {
        return app(Compiler::class)
            ->for($registration)
            ->render($this->message);
    }
}
